class PhonemError(Exception):
    def __init__(self, target_phonem):
        self.target_phonem = target_phonem

    def __str__(self):
        return f'Phonem "{self.target_phonem.transcription}" from "{self.target_phonem.word.original_word}" not found.'


class TokenAmbiguityError(Exception):
    def __init__(self, token):
        self.token = token

    def __str__(self):
        return f'Word "{self.token}" has multiple readings. Please choose a different word.'


class Interruption(Exception):
    def __init__(self, callback):
        self.callback = callback

    def __str__(self):
        return f' Callback "{self.callback.__name_}" returned True'



import sentence_mixing.logic.parameters as params
import sentence_mixing.logic.randomizer as rnd


def rating_sequence_by_phonem_length(x):
    return rnd.noise_score(params.RATING_LENGTH_SAME_PHONEM) * x


def rating_word_by_phonem_length(x):
    return rnd.noise_score(params.RATING_LENGTH_SAME_WORD) * x



import argparse
import os
import random
from sys import argv

import textgrid
import youtube_dl

import sentence_mixing.config as config
import sentence_mixing.logic.analyze as analyze
import sentence_mixing.logic.parameters as params
import sentence_mixing.logic.video_processing as video_processing
import sentence_mixing.model.target as target
from sentence_mixing.serialize import load, save


def prepare_sm_config_file(config_path):
    """Sets the global config dict with the json dict contained in config_path"""
    config.set_config_path(config_path)


def prepare_sm_config_dict(config_dict):
    """Set the global config dict with the dict config_dict"""
    config.set_config_dict(config_dict)


def get_videos(video_urls):
    """
    Builds full SM model from youtube video links.
    The associated youtube videos must have subtitles. Automatic subtitles
    works as well.

    Arguments:
    video_urls -- List[string]: youtube links of the input videos

    Returns:
    A list of Video objects containing the full SM model.
    """

    if not config.is_ready():
        raise Exception(
            "Please, set config file before launching get_videos by calling prepare_sm_config_file or prepare_sm_config_dict"
        )

    (videos,) = load()
    if (
        type(videos) != list
        or len(videos) != len(video_urls)
        or any(map(lambda t: t[0].url != t[1], zip(videos, video_urls)))
    ):
        videos = video_processing.preprocess_and_align(video_urls)
        save(videos)
    return videos


# assuming french for now
def process_sm(
    sentence, videos, seed=params.DEFAULT_SEED, interrupt_callback=None
):
    """
    Reorders the phonems contained in videos model to sound like sentence

    Arguments:
    sentence -- string: the target sentence we want to build from the source
                videos
    videos -- List[model.Video]: contains all the SM model objects from the
              source videos

    Returns:
    A list of combos. A combo is a complete list of choices associating each
    phonem from the target sentence with a real phonem present in a video.
    The combos are stored by relevance. The first combo of the list is the most
    found relevant combination of phonems.
    """

    if not config.is_ready():
        raise Exception(
            "Please, set config file before launching process_sm by calling prepare_sm"
        )

    random.seed(seed)

    # transcribe sentence to pseudo-phonetic string
    target_sentence = target.TargetSentence(sentence)

    combos = analyze.get_n_best_combos(
        target_sentence, videos, interrupt_callback=interrupt_callback
    )
    return combos



import functools

import numpy as np
from scipy import signal

import sentence_mixing.logic.parameters as params


def cross_power_spectral_density_sum(x, y, fs):
    _, Pxy = signal.csd(x[:, 0], y[:, 0], fs, nperseg=1024)
    return np.log(np.abs(Pxy).sum())


def resample(x, current_rate, new_rate):
    return signal.resample(x, int(new_rate * x.shape[0] / current_rate))


def resample_highest_rate(data):
    rate = max(data, key=lambda x: x[0])[0]
    data = [(rate, resample(x, r, rate)) for r, x in data]
    return data


@functools.lru_cache(maxsize=None)
def get_normalized_rms(audio_phonem):
    rate, data = audio_phonem.get_wave()
    return np.average(
        np.sqrt(np.average((data / (1 << 15)) ** 2, axis=-2)), axis=-1
    )


def rate_silence(audio_phonem):
    """
    Compute a silence indicator between 0 (silent) and 1 (saturated).
    """
    average_amplitude = (
        1 - get_normalized_rms(audio_phonem)
    ) ** params.SILENCE_POWER
    return average_amplitude


def rate_amplitude_similarity(previous_audio_segs, audio_seg):
    amplitudes = np.array([get_normalized_rms(w) for w in previous_audio_segs])
    powers = np.arange(
        1, 1 + min(params.AMPLITUDE_STEPS_BACK, len(previous_audio_segs)),
    )
    current_amplitude = get_normalized_rms(audio_seg)
    return float(np.sum(np.abs(amplitudes - current_amplitude) ** powers))


def _soft_rectangle_function(x, a, b, alpha):
    return (np.tanh(alpha * (x - a)) - np.tanh(alpha * (x - b))) / 2


def rate_duration(audio_seg, inf, sup):
    return (
        _soft_rectangle_function(
            audio_seg.get_length(), inf, sup, params.ALPHA
        )
        * 2
        - 1
    )



import os

from moviepy.video.io.VideoFileClip import VideoFileClip
from moviepy.video.compositing.concatenate import concatenate_videoclips


def create_video_file(total_phonems, path, logger="bar"):
    clips = [phonem.get_video_clip() for phonem in total_phonems]
    concatenate_videoclips(clips).write_videofile(path, logger=logger)



import functools
import random

from sentence_mixing.model.association import association_builder

VIDEOS = None


def set_videos(videos):
    """Sets global variable VIDEOS and allows to use related functions"""

    global VIDEOS
    VIDEOS = videos


@functools.lru_cache(maxsize=1)
def get_all_audio_phonems():
    """
    Retrieve all the audio phonems from the video corpus
    Rases exception if video was not set before
    """

    global VIDEOS
    if VIDEOS is None:
        raise Exception(
            "Use set_videos(videos) before running get_all_audio_phonems()"
        )

    audio_phonems = [
        p
        for v in VIDEOS
        for s in v.subtitles
        for w in s.words
        for p in w.phonems
    ]
    random.shuffle(audio_phonems)
    return audio_phonems


@functools.lru_cache(maxsize=1)
def get_transcription_dict_audio_phonem():
    """
    Builds a dictionary associating each possible phonem transcriptions with a list containing
    every audio_phonem having this transcription.
    """

    transcription_audio_phonem_dict = dict()

    for phonem in get_all_audio_phonems():
        if phonem.transcription not in transcription_audio_phonem_dict:
            transcription_audio_phonem_dict[phonem.transcription] = [phonem]
        else:
            transcription_audio_phonem_dict[phonem.transcription].append(
                phonem
            )

    return transcription_audio_phonem_dict


@functools.lru_cache(maxsize=None)
def get_candidates(target_phonem):
    """
    Returns a list of associations where target phonem's transcription and audio phonem's
    transcription are the same.
    This list is sorted by score.
    """

    audio_phonems = get_transcription_dict_audio_phonem()[
        target_phonem.transcription
    ]
    associations = [
        association_builder(target_phonem, audio) for audio in audio_phonems
    ]

    return sorted(
        associations, key=lambda asso: asso.get_total_score(), reverse=True
    )



import random

import sentence_mixing.logic.parameters as params


def noise_score(base_score, sigma=None):
    """
    Adds a gaussian noise on a given score

    Arguments:
    base_score - mean of gaussian score
    sigma - variance of gaussian score. Default is 1/5 of base_score
    """

    if not sigma:
        sigma = base_score * 0.2

    return random.gauss(base_score, sigma)


def random_basic_score():
    """Returns a random basic score following an gaussian distribution"""

    return random.gauss(0, params.RANDOM_SPAN)



import functools
import os
import re

import config

import sentence_mixing.logic.text_parser as tp


@functools.lru_cache(maxsize=1)
def get_speech_to_text_dict():
    """Retrieves the dictionary and parses it to a python dict"""

    dict_path = config.get_property("stt_tmp_dict_path")

    if not os.path.exists(dict_path):
        raise FileNotFoundError(
            "Temporary dictionary not found. Please run generate_compatible_dictionary()"
        )

    # Opens the dictionary file and puts it in a dict
    with open(dict_path) as f:
        phonem_dict = dict()
        previous = None
        for line in f:
            split = line.split(maxsplit=1)
            k, v = split[0], split[1]
            if k == previous:
                phonem_dict[k] = None
            else:
                phonem_dict[k] = v
            previous = k

    return phonem_dict


def generate_compatible_dictionary():
    """Leans speech to text dictionary to keep only words present in SM dictionary"""

    def filter_line(line):
        """Tokenizes the word and checks if it is present in original dict"""

        word = line.split()[0]
        token = word.split("(")[0].upper()

        if token[0] == "-":
            token = token[1:]

        return token in tp.get_dict()

    stt_dict = open(config.get_property("stt_full_dict_path"))

    # Reads dictionary
    stt_dict = stt_dict.readlines()

    # Only keep words e can find in the original dictionary
    stt_dict = list(filter(filter_line, stt_dict))

    tmp_dict_file = config.get_property("stt_tmp_dict_path")
    try:
        os.remove(tmp_dict_file)
    except FileNotFoundError:
        pass

    with open(tmp_dict_file, "w") as filehandle:
        for line in stt_dict:
            filehandle.write(line)


def sentence_to_phonem_list(sentence):
    sentence = sentence.lower()

    # Removing punctuation symbols
    sentence = re.sub(r"\s*([^\s\w'])\s*", "", sentence)

    # Isolating apostrophes
    sentence = re.sub(r"(\w')\s*", "\\1 ", sentence)

    stt_dict = get_speech_to_text_dict()

    phonems = [
        phonem
        for word in sentence.split()
        for phonem in stt_dict[word].split()
    ]

    return phonems



import functools
import re
from itertools import groupby

from num2words import num2words

import sentence_mixing.config as config
from sentence_mixing.model.exceptions import TokenAmbiguityError


@functools.lru_cache(maxsize=None)
def get_dict():
    """Retrieves the dictionary and parses it to a python dict"""

    dict_path = config.get_property("dict_path")

    # Opens the dictionary file and puts it in a dict
    with open(dict_path) as f:
        phonem_dict = dict()
        previous = None
        for line in f:
            split = line.split(maxsplit=1)
            k, v = split[0], split[1]
            if k == previous:
                phonem_dict[k] = None
            else:
                phonem_dict[k] = v
            previous = k

    return phonem_dict


def get_dict_entry(token):
    """Retrieves an entry from the dictionary

    Raises
    ------
    TokenAmbiguityError
        If the token has multiple pronunciations.
    """

    w = get_dict()[token]
    if w is None:
        raise TokenAmbiguityError(token)

    return w.split()


@functools.lru_cache(maxsize=None)
def get_consonant_vowel_dict():
    """
    Retrieves the consonant-vowel dictionary and parses it to a reverse python dict
    Python dict structure: phonem -> phonem_category

    Three categories available: CONSONANT, VOWEL, SPACE
    """

    dict_path = config.get_property("dict_consonant_vowel_path")

    with open(dict_path) as f:
        consonant_vowel_dict = dict(
            [
                [elem, line.split()[0]]
                for line in f
                for elem in line.split()[1:]
            ]
        )

    return consonant_vowel_dict


def are_token_homophones(token0, token1):
    """
    Determines if two token refers to homophones
    For example, TROUVE and TROUVES are homophones, because they are associated to the same phonems
    """

    if token0 == token1:
        return True

    if from_token_to_phonem(token0) == from_token_to_phonem(token1):
        return True

    return False


@functools.lru_cache(maxsize=None)
def get_all_phonems():
    """Retrieves a list of all phonems relying on the consonant-vowel dictionary"""

    with open(config.get_property("dict_consonant_vowel_path")) as f:
        return [phonem for line in f for phonem in line.split()[1:]]


def transform_numbers(text):
    """Substitutes all numbers to plain text in given text"""

    text = re.sub(
        r"(\d+)",
        lambda x: num2words(x.group(1), lang=config.get_property("lang")),
        text,
    )
    return text


def split_text(text):
    """
    Formats the given text to isolate words by:
    - transforming numbers to plain text
    - isolating the punctuation symbols with spaces before and/or after
    - splits on whitespaces
    """

    text = transform_numbers(text)
    text = re.sub(r"(\w')\s*", "\\1 ", text)
    text = re.sub(r"\s*([^\s\w'])\s*", " \\1 ", text)
    return text.split()


def from_word_to_token(word):
    """
    Formats a word to turn it into a token:
    - handles special symbols such as dots
    - puts text in uppercase
    """

    punctuation = [".", "?", "!", ";", ":"]
    if word in punctuation:
        return "SP"

    blank = ["", ","]
    if word in blank:
        return "<BLANK>"

    if not word.replace("'", "a").isalnum():
        return "<TRASH>"

    return word.upper()


def from_token_to_phonem(token):
    """Returns a phonem transcription list corresponding to a given word token

    Raises
    ------
    TokenAmbiguityError
        If the token has multiple pronunciations.
    """

    if token == "SP":
        return ["sp"]

    if token == "<BLANK>":
        return ["sp"]

    if token == "<TRASH>":
        return None

    return get_dict_entry(token)



"""This module contains all the parameters used to compute the combos"""

DEFAULT_SEED = 0

# Step 1

MINIMAL_PHONEM_LENGTH = 0.1
MAXIMAL_MINIMAL_PHONEM_LENGH_MALUS = 1000

MAXIMAL_CONSONANT_LENGTH = 0.25
MAXIMAL_MAXIMAL_CONSONANT_LENGTH_MALUS = 1000

MAXIMAL_VOWEL_LENGTH = 0.5
MAXIMAL_MAXIMAL_VOWEL_LENGTH_MALUS = 1000

# Step 2

RATING_LENGTH_SAME_PHONEM = 80
RATING_LENGTH_SAME_WORD = 100
SILENCE_POWER = 2
SCORE_SAME_TRANSCRIPTION = 200
SCORE_SILENCE_AMPLITUDE = 200
SCORE_DURATION = 400

# Step 3

RATING_SPECTRAL_SIMILARITY = 50
RATING_AMPLITUDE_DIFFERENCE = 500
AMPLITUDE_STEPS_BACK = 4

# Other

NODES = 1 << 12
ALPHA = 1
SCORE_THRESHOLD = 50
RIGHT_PRIVILEGE = 0.8

SCORE_SAME_AUDIO_WORD = 200
RANDOM_SPAN = 50

START_MODIF = 1.0
RATE_POWER = 1.1

ANALYSE_CHUNK_SIZE = 10



import itertools

import sentence_mixing.logic.text_parser as parser
from sentence_mixing.model.abstract import Phonem, Sentence, Word


class TargetSentence(Sentence):
    """Represents the sentence we want to form from the video"""

    def __init__(self, original_text):
        """Asigns the fields and then generates the list of associated TargetWords"""

        Sentence.__init__(self, original_text)

        original_words = parser.split_text(self.original_text)

        interleaved_words = [TargetWord(self, ow) for ow in original_words]
        self.set_words(interleaved_words)

    def next_in_seq(self):
        return None

    def previous_in_seq(self):
        return None


class TargetWord(Word):
    """Represents words inside the TargetSentence"""

    def __init__(self, sentence, original_word):
        """Asigns the fields and then generates the lisr of associated TargetPhonems"""

        token = parser.from_word_to_token(original_word)

        Word.__init__(self, TargetSentence, sentence, token, original_word)

        if self.token != "<TRASH>":
            transcriptions = parser.from_token_to_phonem(self.token)
            self.set_phonems(
                TargetPhonem(self, transcription)
                for transcription in transcriptions
            )


class TargetPhonem(Phonem):
    """Represents phonems in a word"""

    def __init__(self, word, transcription):
        Phonem.__init__(self, TargetWord, word, transcription)



from itertools import groupby

from sentence_mixing.model.abstract import Phonem

BORDERS = {
    (True, True, True, True): "┼",
    (True, True, True, False): "├",
    (True, True, False, True): "┴",
    (True, True, False, False): "└",
    (True, False, True, True): "┤",
    (True, False, True, False): "│",
    (True, False, False, True): "┘",
    (True, False, False, False): "╵",
    (False, True, True, True): "┬",
    (False, True, True, False): "┌",
    (False, True, False, True): "─",
    (False, True, False, False): "╶",
    (False, False, True, True): "┐",
    (False, False, True, False): "╷",
    (False, False, False, True): "╴",
    (False, False, False, False): " ",
}

RATING_STEP = [1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 0]
RATINGS = [
    ("step_1_random", "Random"),
    ("step_1_duration", "Duration"),
    ("step_2_audio_score", "Wave ampl"),
    ("step_2_same_word", "Homophone"),
    ("step_2_word_sequence", "Word seq"),
    ("step_2_phonem_sequence_backward", "Word seq back"),
    ("step_2_phonem_sequence", "Ph sequence"),
    ("step_2_same_phonem", "Transcri"),
    ("step_3_audio_spectral", "Wave freq"),
    ("step_3_audio_amplitude", "Wave amplitude"),
    ("step_3_same_word_previous_phonems", "Previous"),
    ("total", "Total"),
]


def center_in_string(s, n, fill=" "):
    assert len(s) <= n
    left = True
    while len(s) < n:
        if left:
            s = fill + s
        else:
            s += fill
        left = not left
    return s


def combo_displayer(combo):
    phonems = [ch.association.audio_phonem for ch in combo.get_choices()]
    phonems = list(
        zip(
            phonems,
            [
                (
                    ch.association.target_phonem.word,
                    [ch.get_split_score()[k] for k, _ in RATINGS[:-1]],
                )
                for ch in combo.get_choices()
            ],
        )
    )
    phonems_transcriptions = [p.transcription for p, _ in phonems]
    phonems_scores = [score for _, (_, score) in phonems]

    segments_s = [f"{round(p.start, 2)}-{round(p.end, 2)}" for p, _ in phonems]

    COL = BORDERS[(True, False, True, False)]
    LINE = BORDERS[(False, True, False, True)]
    CROSS = BORDERS[tuple(4 * [True])]

    legend_size = max(map(lambda x: len(x[1]), RATINGS[:-1])) + 2
    padding = " " * (legend_size + 1)

    strings = []
    strings.append(COL.join(segments_s))
    strings.append(CROSS.join([len(seg) * LINE for seg in segments_s]))

    for i, _ in enumerate(RATINGS[:-1]):
        strings.append(
            COL.join(
                [
                    center_in_string(str(round(score[i], 2)), len(s))
                    for s, score in zip(segments_s, phonems_scores)
                ]
            )
        )
    strings.append(
        COL.join(
            [
                center_in_string(str(round(sum(score), 2)), len(s))
                for s, score in zip(segments_s, phonems_scores)
            ]
        )
    )
    strings.append(
        COL.join(
            [
                center_in_string(p, len(s))
                for s, p in zip(segments_s, phonems_transcriptions)
            ]
        )
    )
    strings.append(
        strings[1].replace(CROSS, BORDERS[(False, True, True, True)])
    )

    len_by_ws = [
        (w, sum(len(s) + 1 for s, _ in g) - 1)
        for w, g in groupby(zip(segments_s, phonems), key=lambda x: x[1][1][0])
    ]
    strings.append(
        COL.join([center_in_string(w.token, l) for w, l in len_by_ws])
    )
    strings.append("".join([CROSS if c == COL else LINE for c in strings[-1]]))
    strings.append(
        COL.join([center_in_string(w.original_word, l) for w, l in len_by_ws])
    )
    strings.append(
        strings[-2].replace(CROSS, BORDERS[(False, True, True, True)])
    )
    strings.insert(
        0,
        strings[-1].replace(
            BORDERS[(False, True, True, True)],
            BORDERS[(True, True, False, True)],
        ),
    )

    length = 5 + len(RATING_STEP)

    strings.reverse()
    strings = [
        (
            str(RATING_STEP[length - i])
            + RATINGS[length - i][1].rjust(legend_size)
            if 5 < i <= length
            else padding
        )
        + BORDERS[(i != 0, s[0] == LINE, i + 1 != len(strings), False)]
        + s
        + BORDERS[(i != 0, False, i + 1 != len(strings), s[0] == LINE)]
        for i, s in enumerate(strings)
    ]
    return "\n".join(strings)



import functools
import math

import sentence_mixing.logic.parameters as params
import sentence_mixing.logic.randomizer as rnd
import sentence_mixing.logic.text_parser as tp


def get_malus(length, max_length, max_malus):
    if length > max_length:
        multiplier = math.sqrt(max_malus) / max_length
        return min(((length - max_length) * multiplier) ** 2, max_malus,)
    return 0


@functools.lru_cache(maxsize=None)
def score_length(audio_phonem):
    """
    Assigns a score malus to an audio phonem depending on the length of its audio and its phonem type

    The malus is computed following a quadratic distance to a length limit.

    These limits are more or less severe depending of the type of the phonem (consonant, vowel or
    space)

    Returns: a negative score corresponding to a negative malus
    """

    length = audio_phonem.get_length()

    multiplier = (
        math.sqrt(params.MAXIMAL_MINIMAL_PHONEM_LENGH_MALUS)
        / params.MINIMAL_PHONEM_LENGTH
    )
    malus = (max((params.MINIMAL_PHONEM_LENGTH - length), 0) * multiplier) ** 2

    if audio_phonem.get_type() == "CONSONANT":
        malus += get_malus(
            length,
            params.MAXIMAL_CONSONANT_LENGTH,
            params.MAXIMAL_MAXIMAL_CONSONANT_LENGTH_MALUS,
        )
    elif audio_phonem.get_type() == "VOWEL":
        malus += get_malus(
            length,
            params.MAXIMAL_VOWEL_LENGTH,
            params.MAXIMAL_MAXIMAL_VOWEL_LENGTH_MALUS,
        )
    elif audio_phonem.get_type() == "SPACE":
        pass
    return -malus



import base64
from functools import lru_cache

import numpy as np
import webvtt
from moviepy.video.io.VideoFileClip import VideoFileClip
from scipy.io import wavfile

from sentence_mixing.model.audio import SubtitleLine


class Video:
    def __init__(self, url, base_path, subtitles_extension):
        self.url = url
        self._base_path = base_path
        self._subtitles_extension = subtitles_extension
        self.extension = None
        self.subtitles = []
        self.index_subtitles = {}

    def _get_audio_path(self):
        return self._base_path + ".wav"

    def _get_video_path(self):
        assert self.extension is not None
        return self._base_path + "." + self.extension

    def get_subtitle_path(self):
        return self._base_path + self._subtitles_extension

    @lru_cache(maxsize=None)
    def get_audio_wave(self):
        rate, data = wavfile.read(self._get_audio_path())
        if len(np.shape(data)) == 1 or data.shape[1] == 1:
            data = np.stack((data, data), axis=1)
        return (rate, data)

    @lru_cache(maxsize=None)
    def get_video(self):
        return VideoFileClip(self._get_video_path())

    def add_subtitle(self, subtitle):
        assert type(subtitle) == SubtitleLine
        self.index_subtitles[subtitle] = len(self.subtitles)
        self.subtitles.append(subtitle)

    def get_index_subtitle(self, subtitle):
        assert (
            self.index_subtitles is not None
            and subtitle in self.index_subtitles
        )
        return self.index_subtitles[subtitle]

    @lru_cache(maxsize=None)
    def get_subtitle_file(self):
        return webvtt.read(self.get_subtitle_path())

    def get_hashed_basename(self):
        hashed_name = base64.b64encode(self._base_path.encode("utf-8"))
        return str(hashed_name, "utf-8")



import numpy as np
from scipy.io import wavfile


def concat_segments(phonems):
    # TODO resample to the highest rate
    segments = [phonem.get_wave()[1] for phonem in phonems]
    return phonems[0].get_wave()[0], np.concatenate(segments)


def concat_wav(path, phonems):
    rate, new_clip = concat_segments(phonems)

    wavfile.write(path, rate, new_clip)



import pickle


def save(*args, name="save.pckl"):
    with open(name, "wb") as f:
        pickle.dump(args, f)


def load(name="save.pckl"):
    try:
        with open(name, "rb") as f:
            return pickle.load(f)
    except Exception:
        return (None,)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sentence_mixing",
    version="1.1.2",
    author="Nicolas BUSSER, Louis MOUHAT",
    author_email="nicolas.busser67@gmail.com",
    description="Sentence mixing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pop123123123/sentence-mixing",
    packages=["sentence_mixing", "sentence_mixing.logic", "sentence_mixing.model", "sentence_mixing.parameter_tuning", "sentence_mixing.video_creator"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "textgrid",
        "youtube_dl",
        "num2words",
        "numpy",
        "scipy",
        "webvtt-py",
        "moviepy",
        "python-Levenshtein",
        "pocketsphinx",
    ],
    python_requires=">=3.6",
)



import functools
import itertools

import sentence_mixing.logic.analyze_step_2 as step_2
import sentence_mixing.logic.audio_analysis as audio_analysis
import sentence_mixing.logic.parameters as params
import sentence_mixing.logic.utils as utils
from sentence_mixing.model.scorable import Scorable


@functools.lru_cache(maxsize=None)
def association_builder(target_phonem, audio_phonem):
    """Cached association constructor"""

    return Association(target_phonem, audio_phonem)


class Association(Scorable):
    """
    Storage class regrouping target and audio phonem and associating a step 2 score
    It also constains a step 2 score, rating the association of phonems.
    """

    def __init__(self, target_phonem, audio_phonem):
        self.target_phonem = target_phonem
        self.audio_phonem = audio_phonem

        self._step_2_same_phonem_score = (
            target_phonem.transcription == audio_phonem.transcription
        ) * params.SCORE_SAME_TRANSCRIPTION

        self._step_2_same_word_score = (
            utils.are_homophones(target_phonem.word, audio_phonem.word)
            * params.SCORE_SAME_AUDIO_WORD
        )

    @property
    def _step_2_word_sequence_score(self):
        """Rates a common sequence of word between target and audio"""

        if (
            self.target_phonem.get_index_in_word()
            == self.audio_phonem.get_index_in_word()
        ):
            return step_2.rating_word_by_phonem_length(
                len(list(self.sequence_dictionary_homophones_phonems()))
            )
        return 0

    @property
    def _step_2_phonem_sequence_score(self):
        """Rates a common sequence of phonem between target and audio"""

        return step_2.rating_sequence_by_phonem_length(
            max(len(list(self.sequence_same_phonems())) - 1, 0)
        )

    @property
    def _step_2_phonem_sequence_backward_score(self):
        """
        Rates a common sequence of phonem between target and audio.
        This rate is computed in the reverse order and is complementary to
        _step_2_phonem_sequence_score.
        """

        return step_2.rating_sequence_by_phonem_length(
            max(len(list(self.sequence_same_phonems(reverse=True))) - 1, 0)
        )

    @property
    def _step_2_audio_score(self):
        """
        Rates the audio content of a phonem.
        For the <BLANK> tokens, checks if the audio_word is silent and also checks its duration.
        """

        score = 0
        if self.target_phonem.word.token == "<BLANK>":
            score = (
                audio_analysis.rate_silence(self.audio_phonem)
                * params.SCORE_SILENCE_AMPLITUDE
            )
            score += (
                audio_analysis.rate_duration(self.audio_phonem, 0.1, 0.2)
                * params.SCORE_DURATION
            )
        return score

    def __repr__(self):
        return f"<Association {self.target_phonem, self.audio_phonem}>"

    def _get_split_score(self):

        step_2_scores = {
            "step_2_audio_score": self._step_2_audio_score,
            "step_2_same_phonem": self._step_2_same_phonem_score,
            "step_2_same_word": self._step_2_same_word_score,
            "step_2_word_sequence": self._step_2_word_sequence_score,
            "step_2_phonem_sequence": self._step_2_phonem_sequence_score,
            "step_2_phonem_sequence_backward": self._step_2_phonem_sequence_backward_score,
        }
        return {**self.audio_phonem.get_split_score(), **step_2_scores}

    def sequence_dictionary_homophones_phonems(self):
        """
        This method returns a list of associations containing audio phonems corresponding to
        homophone words.
        We force the skip_force_blank argument to True to force get_sequence to evaluate <BLANK>
        only words.
        """

        return utils.association_sequence_from_words(
            utils.get_sequence_dictionary_homophones(
                self.target_phonem, self.audio_phonem, True
            )
        )

    def sequence_aligner_homophones_phonems(self):
        """
        Retrieves associations corresponding to homophones.
        Two words are declared as homophones if they contain the same phonems as the Montreal audio
        phonem decomposition of the word.
        """

        return utils.association_sequence_from_words(
            utils.get_sequence_aligner_homophones(
                self.target_phonem, self.audio_phonem
            )
        )

    @functools.lru_cache(maxsize=None)
    def sequence_same_phonems(self, reverse=False):
        """
        Returns the sequence of identical phonem between target and audio, starting from self
        association.

        Keyword arguments:
        reverse - determines the direction of checking.
                  if True, checks from the end of a word to the begining.
        """
        return list(
            itertools.takewhile(
                lambda association: association.target_phonem.transcription
                == association.audio_phonem.transcription,
                utils.sequence_association(self, reverse=reverse),
            )
        )

    @functools.lru_cache(maxsize=None)
    def sequence_same_phonems_last_word_truncated(self):
        """
        Returns the sequence of identical phonem between target and audio, starting from self
        association.
        Only takes full words: if the common sequence finishes in the middle of the word, trunks
        it to the end of last full word.

        Example:
        Target: "trique fructifier"
        In audio, you can find: "géométrique frugale"
        self: ['t', 't']

        This function returns phonems associated to "trique", even if "fructifier" and "frugale"
        have the same phonem sequence 'f', 'r', 'u'.
        """

        try:
            target_last_end_phonem_word_index = next(
                i
                for i, asso in reversed(
                    list(enumerate(self.sequence_same_phonems()))
                )
                if asso.target_phonem == asso.target_phonem.word.phonems[-1]
            )
            audio_last_end_phonem_word_index = next(
                i
                for i, asso in reversed(
                    list(enumerate(self.sequence_same_phonems()))
                )
                if asso.audio_phonem == asso.audio_phonem.word.phonems[-1]
            )
            return list(self.sequence_same_phonems())[
                : 1
                + min(
                    target_last_end_phonem_word_index,
                    audio_last_end_phonem_word_index,
                )
            ]
        except StopIteration:
            return []

    @functools.lru_cache(maxsize=None)
    def sequence_same_phonems_first_word_truncated(self):
        """
        Returns the sequence of identical phonem between target and audio, starting from self
        association. Breaks when a new word is reached.
        """

        return list(
            map(
                lambda x: x[1],
                itertools.takewhile(
                    lambda x: x[0] == 0
                    or (
                        x[1].target_phonem.get_index_in_word() > 0
                        and x[1].audio_phonem.get_index_in_word() > 0
                    ),
                    enumerate(self.sequence_same_phonems()),
                ),
            )
        )



import functools


class Scorable:
    """Interface implemented by all the objects that carry a score"""

    @functools.lru_cache(maxsize=None)
    def get_total_score(self):
        """Computes final score for a storable"""

        scores = self.get_split_score()
        return sum(scores.values())

    @functools.lru_cache(maxsize=None)
    def get_split_score(self):
        """Cached and public version"""
        return self._get_split_score()

    def _get_split_score(self):
        """Retrieves dict stored detailed score"""

        raise NotImplementedError()
